-- if player has npc_talk_owner priv AND is owner of this particular npc:
--   chat option: "I am your owner. I have new orders for you.
--   -> enters edit mode
-- when edit_mode has been enabled, the following chat options are added to the options:
--   chat option: "Add new answer/option to this dialog."
--   -> adds a new aswer/option
--   chat option: "That was all. I'm finished with giving you new orders. Remember them!"
--   -> ends edit mode
--

--###
-- Init
--###

-- store if the player is editing a particular NPC; format: yl_speak_up.edit_mode[pname] = npc_id
yl_speak_up.edit_mode = {}

-- changes applied in edit_mode are applied immediately - but not immediately stored to disk
-- (this gives the players a chance to back off in case of unwanted changes)
yl_speak_up.npc_was_changed = {}

-- self (the npc as such) is rarely passed on to any functions; in order to be able to check if
-- the player really owns the npc, we need to have that data available;
-- format: yl_speak_up.npc_owner[ npc_id ] = owner_name
yl_speak_up.npc_owner = {}

-- store the current trade between player and npc in case it gets edited in the meantime
yl_speak_up.trade = {}

function yl_speak_up.init_mob_table()
    return false
end

minetest.register_on_leaveplayer(
    function(player)
        yl_speak_up.speak_to[player:get_player_name()] = nil
        yl_speak_up.edit_mode[player:get_player_name()] = nil
    end
)

minetest.register_on_joinplayer(
    function(player)
        yl_speak_up.speak_to[player:get_player_name()] = nil
        yl_speak_up.edit_mode[player:get_player_name()] = nil
    end
)

--###
-- Debug
--###

yl_speak_up.debug = true

--###
-- Helpers
--###

yl_speak_up.get_number_from_id = function(any_id)
    if(any_id == "d_got_item") then
        return "0"
    end
    return string.split(any_id, "_")[2]
end

local function save_path(n_id)
    return yl_speak_up.worldpath .. yl_speak_up.path .. DIR_DELIM .. n_id .. ".json"
end

yl_speak_up.get_error_message = function()
    local formspec = {
        "formspec_version[3]",
        "size[13.4,8.5]",
        "bgcolor[#FF0000]",
        "label[0.2,0.35;Please save a NPC file first]",
        "button_exit[0.2,7.7;3,0.75;button_back;Back]"
    }

    return table.concat(formspec, "")
end

yl_speak_up.find_next_id = function(t)
    local start_id = 1

    if t == nil then
        return start_id
    end

    local keynum = 1
    for k, _ in pairs(t) do
        local keynum = tonumber(yl_speak_up.get_number_from_id(k))
        if keynum >= start_id then
            start_id = keynum + 1
        end
    end
    return start_id
end

yl_speak_up.sanitize_sort = function(options, value)
    local retval = value

    if value == "" or value == nil or tonumber(value) == nil then
        local temp = 0
        for k, v in pairs(options) do
            if v.o_sort ~= nil then
                if tonumber(v.o_sort) > temp then
                    temp = tonumber(v.o_sort)
                end
            end
        end
        retval = tostring(temp + 1)
    end
    return retval
end

--###
--Load and Save
--###

-- we can't really log changes here in this function because we don't know *what* has been changed
yl_speak_up.save_dialog = function(n_id, dialog)
    if type(n_id) ~= "string" or type(dialog) ~= "table" then
        return false
    end
    local p = save_path(n_id)
    -- save some data (in particular usage of quest variables)
    yl_speak_up.update_stored_npc_data(n_id, dialog)
    local content = minetest.write_json(dialog)
    return minetest.safe_file_write(p, content)
end


yl_speak_up.load_dialog = function(n_id) -- returns the saved dialog
    local p = save_path(n_id)

    local file, err = io.open(p, "r")
    if err then
        return {}
    end
    io.input(file)
    local content = io.read()
    local dialog = minetest.parse_json(content)
    io.close(file)

    if type(dialog) ~= "table" then
        dialog = {}
    end

    return dialog
end

-- used by staff and input_inital_config
yl_speak_up.fields_to_dialog = function(pname, fields)
    local n_id = yl_speak_up.speak_to[pname].n_id
    local dialog = yl_speak_up.load_dialog(n_id)
    local save_d_id = ""

    if next(dialog) == nil then -- No file found. Let's create the basic values
        dialog = {}
        dialog.n_dialogs = {}
    end

    if dialog.n_dialogs == nil or next(dialog.n_dialogs) == nil then --No dialogs found. Let's make a table
        dialog.n_dialogs = {}
    end

    if fields.d_text ~= "" then -- If there is dialog text, then save new or old dialog
        if fields.d_id == yl_speak_up.text_new_dialog_id then --New dialog --
            -- Find highest d_id and increase by 1
            save_d_id = "d_" .. yl_speak_up.find_next_id(dialog.n_dialogs)

            -- Initialize empty dialog
            dialog.n_dialogs[save_d_id] = {}
        else -- Already existing dialog
            save_d_id = fields.d_id
        end
        -- Change dialog
        dialog.n_dialogs[save_d_id].d_id = save_d_id
        dialog.n_dialogs[save_d_id].d_type = "text"
        dialog.n_dialogs[save_d_id].d_text = fields.d_text
        dialog.n_dialogs[save_d_id].d_sort = fields.d_sort
    end

    --Context
    yl_speak_up.speak_to[pname].d_id = save_d_id

    -- Just in case the NPC vlaues where changed or set
    dialog.n_id = n_id
    dialog.n_description = fields.n_description
    dialog.n_npc = fields.n_npc

    dialog.npc_owner = fields.npc_owner

    return dialog
end

yl_speak_up.delete_dialog = function(n_id, d_id)
    if d_id == yl_speak_up.text_new_dialog_id then
        return false
    end -- We don't delete "New dialog"

    local dialog = yl_speak_up.load_dialog(n_id)

    dialog.n_dialogs[d_id] = nil

    yl_speak_up.save_dialog(n_id, dialog)
end


--###
--Formspecs
--###

-- get formspecs

-- talk

-- receive fields


-- talk

-- helper function
-- the option to override next_id and provide a value is needed when a new dialog was
-- added, then edited, and then discarded; it's still needed after that, but has to
-- be reset to empty state (wasn't stored before)
yl_speak_up.add_new_dialog = function(dialog, pname, next_id)
	if(not(next_id)) then
		next_id = yl_speak_up.find_next_id(dialog.n_dialogs)
	end
	local future_d_id = "d_" .. next_id
	-- Initialize empty dialog
	dialog.n_dialogs[future_d_id] = {
		d_id = future_d_id,
		d_type = "text",
		d_text = "",
		d_sort = next_id
		}
	-- store that there have been changes to this npc
	-- (better ask only when the new dialog is changed)
--	table.insert(yl_speak_up.npc_was_changed[ yl_speak_up.edit_mode[pname] ],
--		"Dialog "..future_d_id..": New dialog added.")
	return future_d_id
end


-- add a new result to option o_id of dialog d_id
yl_speak_up.add_new_result = function(dialog, d_id, o_id)
	if(not(dialog) or not(dialog.n_dialogs) or not(dialog.n_dialogs[d_id])
	  or not(dialog.n_dialogs[d_id].d_options) or not(dialog.n_dialogs[d_id].d_options[o_id])) then
		return
	end
	-- create a new result (first the id, then the actual result)
	local future_r_id = "r_" .. yl_speak_up.find_next_id(dialog.n_dialogs[d_id].d_options[o_id].o_results)
	if future_r_id == "r_1" then
		dialog.n_dialogs[d_id].d_options[o_id].o_results = {}
	end
	dialog.n_dialogs[d_id].d_options[o_id].o_results[future_r_id] = {}
	return future_r_id
end


-- this is useful for result types that can exist only once per option
-- (apart from editing with the staff);
-- examples: "dialog" and "trade";
-- returns tue r_id or nil if no result of that type has been found
yl_speak_up.get_result_id_by_type = function(dialog, d_id, o_id, result_type)
	if(not(dialog) or not(dialog.n_dialogs) or not(dialog.n_dialogs[d_id])
	  or not(dialog.n_dialogs[d_id].d_options) or not(dialog.n_dialogs[d_id].d_options[o_id])) then
		return
	end
	local results = dialog.n_dialogs[d_id].d_options[o_id].o_results
	if(not(results)) then
		return
	end
	for k, v in pairs(results) do
		if(v.r_type == result_type) then
			return k
		end
	end
end


-- helper function for sorting options/answers using options[o_id].o_sort
-- (or dialogs by d_sort)
yl_speak_up.get_sorted_options = function(options, sort_by)
	local sorted_list = {}
	for k,v in pairs(options) do
		table.insert(sorted_list, k)
	end
	table.sort(sorted_list,
		function(a,b)
			if(not(options[a][sort_by])) then
				return false
			elseif(not(options[b][sort_by])) then
				return true
			-- sadly not all entries are numeric
			elseif(tonumber(options[a][sort_by]) and tonumber(options[b][sort_by])) then
				return (tonumber(options[a][sort_by]) < tonumber(options[b][sort_by]))
			-- numbers have a higher priority
			elseif(tonumber(options[a][sort_by])) then
				return true
			elseif(tonumber(options[b][sort_by])) then
				return false
			-- if the value is the same: sort by index
			elseif(options[a][sort_by] == options[b][sort_by]) then
				return (a < b)
			else
				return (options[a][sort_by] < options[b][sort_by])
			end
		end
	)
	return sorted_list
end


-- simple sort of keys of a table numericly;
-- this is not efficient - but that doesn't matter: the lists are small and
-- it is only executed when configuring an NPC
-- simple: if the parameter is true, the keys will just be sorted (i.e. player names) - which is
-- 	not enough for d_<nr>, o_<nr> etc. (which need more care when sorting)
yl_speak_up.sort_keys = function(t, simple)
	local keys = {}
	for k, v in pairs(t) do
		-- add a prefix so that p_2 ends up before p_10
		if(not(simple) and string.len(k) == 3) then
			k = "a"..k
		end
		table.insert(keys, k)
	end
	table.sort(keys)
	if(simple) then
		return keys
	end
	for i,k in ipairs(keys) do
		-- avoid cutting the single a from a_1 (action 1)
		if(k and string.sub(k, 1, 1) == "a" and string.sub(k, 2, 2) ~= "_") then
			-- remove the leading blank
			keys[i] = string.sub(k, 2)
		end
	end
	return keys
end


-- helper function for yl_speak_up.edit_mode_apply_changes;
-- makes sure the new dialog (and a result/effect "dialog" for each option) exist
yl_speak_up.prepare_new_dialog_for_option = function(dialog, pname, n_id, d_id, o_id,target_dialog,o_results)
	-- is there a result/effect of the type "dialog" already? else use a fallback
	local result = {r_value = "-default-"}
	if(o_results) then
		for kr, vr in pairs(o_results) do
			if( vr.r_type == "dialog" ) then
				result = vr
			end
		end
	end
	-- this may also point to a new dialog
	if(target_dialog == yl_speak_up.text_new_dialog_id) then
		-- create a new dialog and show it as new target dialog - but do not display
		-- this dialog directly (the player may follow the -> button)
		target_dialog = yl_speak_up.add_new_dialog(dialog, pname, nil)
	elseif( result.r_value and result.r_value == target_dialog) then
		-- no problem - the right dialog is set already
		return target_dialog
	end
	-- store that a new option has been added to this dialog
	table.insert(yl_speak_up.npc_was_changed[ n_id ],
		"Dialog "..d_id..": The target dialog for option "..
		tostring(o_id).." was changed from "..
		tostring(result.r_value or "-default-").." to "..
		tostring(target_dialog)..".")
	-- does the result/effect of type "dialog" exist already? then we're done
	if(result.r_type and result.r_type == "dialog") then
		-- actually change the target dialog
		result.r_value = target_dialog
		return target_dialog
	end
	-- create a new result (first the id, then the actual result)
	local future_r_id = yl_speak_up.add_new_result(dialog, d_id, o_id)
	-- actually store the new result
	dialog.n_dialogs[d_id].d_options[o_id].o_results[future_r_id] = {
		r_id = future_r_id,
		r_type = "dialog",
		r_value = target_dialog}
	return target_dialog
end


-- helper function for formspec "yl_speak_up:talk" *and* formspec "yl_speak_up:edit_option_dialog"
-- when a parameter was changed in edit mode;
-- this is called when the player is in edit_mode (editing the NPC);
-- the function checks if the player has changed any parameters
-- Parameters:
--    pname    player name
--    fields   the fields returned from the formspec
-- Returns:
--    result   table with information about what was added
--             (for now, only result.show_next_option is of intrest in the option edit menu)
yl_speak_up.edit_mode_apply_changes = function(pname, fields)
	local n_id = yl_speak_up.edit_mode[pname]
	local d_id = yl_speak_up.speak_to[pname].d_id
	local dialog = yl_speak_up.speak_to[pname].dialog

	-- this way we can store the actual changes and present them to the player for saving
	if(not(yl_speak_up.npc_was_changed[ n_id ])) then
		yl_speak_up.npc_was_changed[ n_id ] = {}
	end


	-- nothing to do if that dialog does not exist
	if(not(d_id) or not(dialog.n_dialogs) or not(dialog.n_dialogs[ d_id ])) then
		return
	end

	-- allow owner to mute/unmute npc (would be bad if players can already see what is going
	-- to happen while the owner creates a long quest)
	-- mute/unmute gets logged in the function and does not need extra log entries
	local obj = yl_speak_up.speak_to[pname].obj
	if(fields.mute_npc and obj) then
		yl_speak_up.set_muted(pname, obj, true)
	elseif(fields.un_mute_npc and obj) then
		yl_speak_up.set_muted(pname, obj, false)
	end


	-- new options etc. may be added; store these IDs so that we can switch to the right target
	local result = {}

	-- make this the first dialog shown when starting a conversation
	if(fields.make_first_option) then
		-- check which dialog(s) previously had the highest priority and change thsoe
		for k, v in pairs(dialog.n_dialogs) do
			if(v and v.d_sort and (v.d_sort=="0" or v.d_sort==0)) then
				-- try to derive a sensible future sort priority from the key:
				-- here we make use of the d_<nr> pattern; but even if that fails to yield
				-- a number, the sort function will later be able to deal with it anyway
				local new_priority = string.sub(k, 3)
				dialog.n_dialogs[ k ].d_sort = new_priority
			end
		end
		-- actually make this the chat with the highest priority
		dialog.n_dialogs[ d_id ].d_sort = "0"
		-- this is not immediately saved, even though the changes to the previous dialog with
		-- the highest priority cannot be automaticly undone (but as long as it is not saved,
		-- it really does not matter; and when saving, the player has to take some care)
		table.insert(yl_speak_up.npc_was_changed[ n_id ],
			"Dialog "..d_id..": Turned into new start dialog.")
	end

	-- detect changes to d_text: text of the dialog (what the npc is saying)
	-- (only happens in dialog edit menu)
	if(fields.d_text and dialog.n_dialogs[ d_id ].d_text ~= fields.d_text) then
		-- store that there have been changes to this npc
		table.insert(yl_speak_up.npc_was_changed[ n_id ],
			"Dialog "..d_id..": d_text (what the NPC says) was changed from \""..
			tostring( dialog.n_dialogs[ d_id ].d_text)..
			"\" to \""..tostring(fields.d_text).."\".")
		-- actually change the text - but do not save to disk yet
		dialog.n_dialogs[ d_id ].d_text = fields.d_text
	end

	-- add a new option/answer
	if(fields[ "add_option"]) then
		-- count the options/answers so that the limit is not exceeded
		local anz_options = 0
		local future_o_id = "o_" .. yl_speak_up.find_next_id(dialog.n_dialogs[d_id].d_options)
		if dialog.n_dialogs[d_id].d_options == nil then
			dialog.n_dialogs[d_id].d_options = {}
		else
			local sorted_list = yl_speak_up.get_sorted_options(dialog.n_dialogs[d_id].d_options, "o_sort")
			anz_options = #sorted_list
		end
		-- we don't want an infinite amount of answers per dialog
		if(anz_options >= yl_speak_up.max_number_of_options_per_dialog) then
			-- this is already checked earlier on and the button only shown if
			-- options can be added; so this can reamin a chat message
			minetest.chat_send_player(pname, "Sorry. Only "..
				tostring(yl_speak_up.max_number_of_options_per_dialog)..
				" options/answers are allowed per dialog.")
			fields.add_option = nil
		else
			dialog.n_dialogs[d_id].d_options[future_o_id] = {
				o_id = future_o_id,
				o_hide_when_prerequisites_not_met = "false",
				o_grey_when_prerequisites_not_met = "false",
				o_sort = -1,
				o_text_when_prerequisites_not_met = "",
				o_text_when_prerequisites_met = "",
				}
			-- necessary in order for it to work
			local s = yl_speak_up.sanitize_sort(dialog.n_dialogs[d_id].d_options, yl_speak_up.speak_to[pname].o_sort)
			dialog.n_dialogs[d_id].d_options[future_o_id].o_sort = s
			table.insert(yl_speak_up.npc_was_changed[ n_id ],
				"Dialog "..d_id..": Added new option/answer "..future_o_id..".")
			-- if this is selected in the options edit menu, we want to move straight on to the new option
			result["show_next_option"] = future_o_id
		end
	end

	if(fields[ "del_option"] and fields.o_id and dialog.n_dialogs[d_id].d_options[fields.o_id]) then
		local o_id = fields.o_id
		-- which dialog to show instead of the deleted one?
		local next_o_id = o_id
		local sorted_list = yl_speak_up.get_sorted_options(dialog.n_dialogs[d_id].d_options, "o_sort")
		for i, o in ipairs(sorted_list) do
			if(o == o_id and sorted_list[ i+1 ]) then
				next_o_id = sorted_list[ i+1 ]
			elseif(o == o_id and sorted_list[ i-1 ]) then
				next_o_id = sorted_list[ i-1 ]
			end
		end
		table.insert(yl_speak_up.npc_was_changed[ n_id ],
			"Dialog "..d_id..": Option "..tostring(o_id).." deleted.")
		-- actually delete the dialog
		dialog.n_dialogs[d_id].d_options[o_id] = nil
		-- the current dialog is deleted; we need to show another one
		result["show_next_option"] = next_o_id
		-- after deleting the entry, all previous/further changes to it are kind of unintresting
		return result
	end

	-- ignore entries to o_sort if they are not a number
	if(fields[ "edit_option_o_sort"]
	  and tonumber(fields[ "edit_option_o_sort"])
	  and fields.o_id and dialog.n_dialogs[d_id].d_options[fields.o_id]) then
		local o_id = fields.o_id
		local new_nr = tonumber(fields[ "edit_option_o_sort"])
		local old_nr = tonumber(dialog.n_dialogs[d_id].d_options[o_id].o_sort)
		-- if the nr is -1 (do not show) then we are done already: nothing to do
		if(old_nr == new_nr) then
		-- -1: do not list as option/answer (but still store and keep it)
		elseif(new_nr == -1 and old_nr ~= -1) then
			dialog.n_dialogs[d_id].d_options[o_id].o_sort = "-1"
			table.insert(yl_speak_up.npc_was_changed[ n_id ],
				"Dialog "..d_id..": Option "..tostring(o_id).." was set to -1 (do not list).")
		else
			-- get the old sorted list
			local sorted_list = yl_speak_up.get_sorted_options(dialog.n_dialogs[d_id].d_options, "o_sort")
			-- negative numbers are not shown
			local entries_shown_list = {}
			for i, o in ipairs(sorted_list) do
				local n = tonumber(dialog.n_dialogs[d_id].d_options[o].o_sort)
				if(n and n > 0 and o ~= o_id) then
					table.insert(entries_shown_list, o)
				end
			end
			-- insert the entry at the new position and let lua do the job
			table.insert(entries_shown_list, new_nr, o_id)
			-- take the indices from that new list as new sort values and store them;
			-- this has the side effect that duplicate entries get sorted out as well
			for i, o in ipairs(entries_shown_list) do
				dialog.n_dialogs[d_id].d_options[o].o_sort = tostring(i)
			end
			-- store that there was a cahnge
			table.insert(yl_speak_up.npc_was_changed[ n_id ],
				"Dialog "..d_id..": Option "..tostring(o_id).." was moved to position "..
				tostring(new_nr)..".")
		end
	end

	-- changes to options are not possible if there are none
	if(dialog.n_dialogs[ d_id ].d_options) then

		-- detect changes to text_option_<o_id>: text for option <o_id>
		for k, v in pairs(dialog.n_dialogs[ d_id ].d_options) do
			if(   fields[ "text_option_"..k ]
			  and fields[ "text_option_"..k ] ~= v.o_text_when_prerequisites_met ) then
				-- store that there have been changes to this npc
				table.insert(yl_speak_up.npc_was_changed[ n_id ],
					"Dialog "..d_id..": The text for option "..tostring(k)..
					" was changed from \""..tostring(v.o_text_when_prerequisites_met)..
					"\" to \""..tostring(fields[ "text_option_"..k]).."\".")
				-- actually change the text of the option
				dialog.n_dialogs[ d_id ].d_options[ k ].o_text_when_prerequisites_met = fields[ "text_option_"..k ]
			end
		end

		-- detect changes to d_id_<o_id>: target dialog for option <o_id>
		for k, v in pairs(dialog.n_dialogs[ d_id ].d_options) do
			if(fields[ "d_id_"..k ]) then
				local new_target_dialog = yl_speak_up.prepare_new_dialog_for_option(
					dialog, pname, n_id, d_id, k, fields[ "d_id_"..k ], v.o_results)
				if(new_target_dialog ~= fields[ "d_id_"..k ]) then
					fields[ "d_id_"..k ] = new_target_dialog
					-- in options edit menu: show this update
					result["show_next_option"] = k
				end
			end
		end
	end

	-- add a new dialog; either via "+" button or "New dialog" in dialog dropdown menu
	-- this has to be done after all the other changes because those (text changes etc.) still
	-- apply to the *old* dialog
	if(fields.show_new_dialog
	  or(fields["d_id"] and fields["d_id"] == yl_speak_up.text_new_dialog_id)) then
		-- create the new dialog and make sure it gets shown
		local d_id = yl_speak_up.add_new_dialog(dialog, pname, nil)
		-- actually show the new dialog
		fields["d_id"] = d_id
		fields["show_new_dialog"] = nil
	end

	-- delete one empty dialog
	if(fields.delete_this_empty_dialog) then
		local anz_options = 0
		-- we need to show a new dialog after this one was deleted
		local new_dialog = d_id
		local sorted_list = yl_speak_up.get_sorted_options(dialog.n_dialogs, "d_sort")
		for i, k in ipairs(sorted_list) do
			-- count the options of this dialog
			if(k == d_id) then
				if(dialog.n_dialogs[d_id].d_options) then
					for o, w in pairs(dialog.n_dialogs[d_id].d_options) do
						anz_options = anz_options + 1
					end
				end
				if(sorted_list[i+1]) then
					new_dialog = sorted_list[i+1]
				elseif(sorted_list[i-1]) then
					new_dialog = sorted_list[i-1]
				end
			end
		end
		-- there needs to be one dialog left after deleting this one,
		if(#sorted_list > 1
		-- this dialog isn't allowed to hold any more options/answers
		  and anz_options == 0
		-- we really found a new dialog to show
		  and new_dialog ~= d_id
		-- and the text needs to be empty
		  and dialog.n_dialogs[ d_id ].d_text == "") then
			-- actually delete this dialog
			dialog.n_dialogs[ d_id ] = nil
			-- ..and store it to disk
			yl_speak_up.delete_dialog(n_id, d_id)
                        yl_speak_up.log_change(pname, n_id,
				"Deleted dialog "..tostring(d_id)..".")
			-- switch to another dialog (this one was deleted after all)
			fields["d_id"] = new_dialog
			fields["show_new_dialog"] = nil
		else
			-- deleting is only possible from the talk menu, and there the delete
			-- button is only shown if the dialog can be deleted; so this can remain
			-- a chat message
			minetest.chat_send_player(pname, "Sorry. This dialog cannot be deleted (yet). "..
				"It is either the only dialog left or has a non-empty text or has at "..
				"least on remaining option/answer.")
		end
	end

	-- not in options edit menu?
	local o_id = fields.o_id
	if(not(o_id)) then
		return result
	end

	local d_option = dialog.n_dialogs[ d_id ].d_options[ o_id ]
	-- change alternate text when preconditions are not met
	-- (only happens in options edit menu)
	if(fields.option_text_not_met and d_option
	  and d_option.o_text_when_prerequisites_not_met ~= fields.option_text_not_met) then
		-- add change to changelog
		table.insert(yl_speak_up.npc_was_changed[ n_id ],
			"Dialog "..d_id..": The alternate text for option "..tostring(o_id)..
			" was changed from \""..
			tostring(d_option.o_text_when_prerequisites_not_met).."\" to \""..
			tostring(fields.option_text_not_met).."\".")
		-- actually change the text of the option
		d_option.o_text_when_prerequisites_not_met = fields.option_text_not_met
	end

	-- toggle autoselection/autoclick of an option
	if(d_option and fields.option_autoanswer and fields.option_autoanswer ~= "") then
		if(d_option.o_autoanswer and d_option.o_autoanswer == 1
		  and fields.option_autoanswer == "by clicking on it") then
			d_option.o_autoanswer = nil
			table.insert(yl_speak_up.npc_was_changed[ n_id ],
				"Dialog "..d_id..": The modus for option "..tostring(o_id)..
				" was changed from \"automaticly\" to \"by clicking on it\".")
		elseif((not(d_option.o_autoanswer) or d_option.o_autoanswer ~= 1)
		  and fields.option_autoanswer == "automaticly") then
			d_option.o_autoanswer = 1
			table.insert(yl_speak_up.npc_was_changed[ n_id ],
				"Dialog "..d_id..": The modus for option "..tostring(o_id)..
				" was changed from \"by clicking on it\" to \"automaticly\".")
		end
	end

	-- handle hide/grey out/show alternate answer
	-- (only happens in options edit menu)
	if(fields.hide_or_grey_or_alternate_answer and d_option) then
		if(fields.hide_or_grey_or_alternate_answer == "..hide this answer."
		  and d_option.o_hide_when_prerequisites_not_met ~= "true") then
			d_option.o_hide_when_prerequisites_not_met = "true"
			d_option.o_grey_when_prerequisites_not_met = "false"
			table.insert(yl_speak_up.npc_was_changed[ n_id ],
				"Dialog "..d_id..": If precondition for option "..tostring(o_id)..
				" is not met, hide option/answer.")
			-- make sure we show this options update next
			result["show_next_option"] = o_id
		elseif(fields.hide_or_grey_or_alternate_answer == "..grey out the following answer:"
		  and d_option.o_grey_when_prerequisites_not_met ~= "true") then
			d_option.o_hide_when_prerequisites_not_met = "false"
			d_option.o_grey_when_prerequisites_not_met = "true"
			table.insert(yl_speak_up.npc_was_changed[ n_id ],
				"Dialog "..d_id..": If precondition for option "..tostring(o_id)..
				" is not met, grey out option/answer.")
			result["show_next_option"] = o_id
		elseif(fields.hide_or_grey_or_alternate_answer == "..display the following alternate answer:"
		  and (d_option.o_hide_when_prerequisites_not_met ~= "false"
		    or d_option.o_grey_when_prerequisites_not_met) ~= "false") then
			d_option.o_hide_when_prerequisites_not_met = "false"
			d_option.o_grey_when_prerequisites_not_met = "false"
			table.insert(yl_speak_up.npc_was_changed[ n_id ],
				"Dialog "..d_id..": If precondition for option "..tostring(o_id)..
				" is not met, show alternate option/answer.")
			result["show_next_option"] = o_id
		end
	end

	-- how many times can the player fail to execute the action successfully?
	if(fields[ "timer_max_attempts_on_failure"]) then
		local field_name = "timer_max_attempts_on_failure"
		local timer_name = "timer_on_failure_"..tostring(d_id).."_"..tostring(o_id)
		if(not(tonumber(fields[ field_name ]))) then
			fields[ field_name ] = 0
		end
		-- make sure the variable exists
		if(yl_speak_up.add_time_based_variable(timer_name)) then
			yl_speak_up.set_variable_metadata(timer_name, nil, "parameter", "max_attempts",
							  fields[ field_name ])
		end
	end
	-- ..and how long has the player to wait in order to try again?
	if(fields[ "timer_max_seconds_on_failure"]) then
		local field_name = "timer_max_seconds_on_failure"
		local timer_name = "timer_on_failure_"..tostring(d_id).."_"..tostring(o_id)
		if(not(tonumber(fields[ field_name ]))) then
			fields[ field_name ] = 0
		end
		-- make sure the variable exists
		if(yl_speak_up.add_time_based_variable(timer_name)) then
			yl_speak_up.set_variable_metadata(timer_name, nil, "parameter", "duration",
							  fields[ field_name ])
		end
	end
	if(fields[ "timer_max_seconds_on_success"]) then
		local field_name = "timer_max_seconds_on_success"
		local timer_name = "timer_on_success_"..tostring(d_id).."_"..tostring(o_id)
		if(not(tonumber(fields[ field_name ]))) then
			fields[ field_name ] = 0
		end
		-- make sure the variable exists
		if(yl_speak_up.add_time_based_variable(timer_name)) then
			yl_speak_up.set_variable_metadata(timer_name, nil, "parameter", "duration",
							  fields[ field_name ])
		end
	end
	-- currently only contains result["show_new_option"] (which is needed for options edit menu)
	return result
end
-- end of yl_speak_up.edit_mode_apply_changes



-- identify multiple results that lead to target dialogs
yl_speak_up.check_for_disambigous_results = function(n_id, pname)
	local errors_found = false
	local dialog = yl_speak_up.load_dialog(n_id)
	-- nothing defined yet - nothing to repair
	if(not(dialog.n_dialogs)) then
		return
	end
	-- iterate over all dialogs
	for d_id, d in pairs(dialog.n_dialogs) do
		if(d_id and d and d.d_options) then
			-- iterate over all options
			for o_id, o in pairs(d.d_options) do
				if(o_id and o and o.o_results) then
					local dialog_results = {}
					-- iterate over all results
					for r_id, r in pairs(o.o_results) do
						if(r.r_type == "dialog") then
							table.insert(dialog_results, r_id)
						end
					end
					if(#dialog_results>1) then
						minetest.chat_send_player(pname, "ERROR: Dialog "..
							tostring(d_id)..", option "..tostring(o_id)..
							", has multiple results of type dialog: "..
							minetest.serialize(dialog_results)..". Please "..
							"let someone with npc_master priv fix that first!")
						errors_found = true
					end
				end
			end
		end
	end
	return errors_found
end


-- Make the NPC talk

--###
-- Mob functions
--###

function yl_speak_up.on_rightclick(self, clicker)
    --local item = clicker:get_wielded_item()
    local name = clicker:get_player_name()

    -- Take the mob only with net or lasso
    if self.owner and self.owner == name then
        if mobs:capture_mob(self, clicker, nil, 100, 100, true, nil) then
            return
        end
    end

    -- protect npc with mobs:protector
    if mobs:protect(self, clicker) then
        return
    end

    -- bring up the dialog options
    if clicker then
        yl_speak_up.talk(self, clicker)
        return
    end
end

function yl_speak_up.on_spawn(self)
    --Let's assign an ID

    local m_talk = yl_speak_up.talk_after_spawn or true
    local m_id = yl_speak_up.number_of_npcs + 1

    yl_speak_up.number_of_npcs = m_id
    yl_speak_up.modstorage:set_int("amount", m_id)

    yl_speak_up.mob_table[m_id] = "yl_speak_up:test_npc"

    self.yl_speak_up = {
        talk = m_talk,
        id = m_id,
        textures = self.textures
    }

    --Let's protect it
    self.protected = true
    self.tamed = true
    self.object:set_armor_groups({immortal = 100})

    minetest.log(
        "action",
        "[MOD] yl_speak_up: NPC with ID n_" ..
            self.yl_speak_up.id .. " spawned at " .. minetest.pos_to_string(self.object:get_pos(), 0)
    )

    --Let's do it only once
    return true
end

function yl_speak_up.after_activate(self, staticdata, def, dtime)
    minetest.log(
        "action",
        "[MOD] yl_speak_up: NPC with ID n_" ..
            self.yl_speak_up.id .. " activated at " .. minetest.pos_to_string(self.object:get_pos(), 0)
    )

    if yl_speak_up.status == 2 then
        self.object:remove()
        return true
    end

    if self.yl_speak_up and self.yl_speak_up.skin then
        local tex = self.yl_speak_up.skin
        self.object:set_properties({textures = {tex[1], tex[2], tex[3], tex[4]}})
    end

    if yl_speak_up.infotext then
        local i_text = ""
        if self.yl_speak_up.npc_name then
            i_text = i_text .. self.yl_speak_up.npc_name .. "\n"
        end
        if self.yl_speak_up.npc_description then
            i_text = i_text .. self.yl_speak_up.npc_description .. "\n"
        end
        i_text = i_text .. yl_speak_up.infotext
        self.object:set_properties({infotext = i_text})
    end

    yl_speak_up.update_nametag(self)

    -- create a detached inventory for the npc and load its inventory
    yl_speak_up.load_npc_inventory("n_"..tostring(self.yl_speak_up.id))
end

function yl_speak_up.talk(self, clicker)

    if not clicker and not clicker:is_player() then
        return
    end
    if not self then
        return
    end
    if not self.yl_speak_up or not self.yl_speak_up.id then
        return
    end

    local npc_id = self.yl_speak_up.id
    local n_id = "n_" .. npc_id

    -- remember whom the npc belongs to (as long as we still have self.owner available for easy access)
    yl_speak_up.npc_owner[ n_id ] = self.owner

    local pname = clicker:get_player_name()
    if not self.yl_speak_up or not self.yl_speak_up.talk or self.yl_speak_up.talk~=true then

	-- show a formspec to other players that this NPC is busy
        if(not(yl_speak_up.may_edit_npc(clicker, n_id))) then
             -- show a formspec so that the player knows that he may come back later
             yl_speak_up.show_fs(player, "msg", {input_to = "yl_spaek_up:ignore", formspec =
		"size[6,2]"..
		"label[1.2,0.0;"..minetest.formspec_escape((self.yl_speak_up.npc_name or "This NPC")..
			" [muted]").."]"..
		"label[0.2,0.5;Sorry! I'm currently busy learning new things.]"..
		"label[0.2,1.0;Please come back later.]"..
		"button_exit[2.5,1.5;1,0.9;ok;Ok]"})
             return
        end
        -- allow the owner to edit (and subsequently unmute) the npc
	minetest.chat_send_player(pname, "This NPC is muted. It will only talk to you.")
    end

    yl_speak_up.speak_to[pname] = {}
    yl_speak_up.speak_to[pname].n_id = n_id -- Memorize which player talks to which NPC
    yl_speak_up.speak_to[pname].dialog = yl_speak_up.load_dialog(n_id) -- Load the dialog and see what we can do with it
    yl_speak_up.speak_to[pname].textures = self.yl_speak_up.textures
    yl_speak_up.speak_to[pname].option_index = 1
    yl_speak_up.speak_to[pname].obj = self.object

    -- is this player explicitly allowed to edit this npc?
    if(yl_speak_up.speak_to[pname].dialog
      and yl_speak_up.speak_to[pname].dialog.n_may_edit
      and yl_speak_up.speak_to[pname].dialog.n_may_edit[pname]
      and minetest.check_player_privs(clicker, {npc_talk_owner=true})) then
	yl_speak_up.speak_to[pname].may_edit_this_npc = true
    end

    yl_speak_up.show_fs(clicker, "talk", {n_id = n_id})

    local dialog = yl_speak_up.speak_to[pname].dialog
    if(not(dialog.trades)) then
       dialog.trades = {}
    end
end


-- mute the npc; either via the appropriate staff or via talking to him
yl_speak_up.set_muted = function(p_name, obj, set_muted)
	if(not(obj)) then
		return
	end
	local luaentity = obj:get_luaentity()
	if(not(luaentity)) then
		return
	end
	local npc = luaentity.yl_speak_up.id
	local npc_name = luaentity.yl_speak_up.npc_name
	-- fallback
	if(not(npc_name)) then
		npc_name = npc
	end
	if(set_muted and luaentity.yl_speak_up.talk) then
		-- the npc is willing to talk
		luaentity.yl_speak_up.talk = false
		yl_speak_up.update_nametag(luaentity)

		minetest.chat_send_player(p_name,"NPC with ID "..npc.." will shut up at pos "..
			minetest.pos_to_string(obj:get_pos(),0).." on command of "..p_name)
		minetest.log("action","[MOD] yl_speak_up: NPC with ID n_"..npc..
			" will shut up at pos "..minetest.pos_to_string(obj:get_pos(),0)..
			" on command of "..p_name)
	elseif(not(set_muted) and not(luaentity.yl_speak_up.talk)) then
		-- mute the npc
		luaentity.yl_speak_up.talk = true
		yl_speak_up.update_nametag(luaentity)

		minetest.chat_send_player(p_name,"NPC with ID "..npc.." will resume speech at pos "..
			minetest.pos_to_string(obj:get_pos(),0).." on command of "..p_name)
		minetest.log("action","[MOD] yl_speak_up: NPC with ID n_"..npc..
			" will resume speech at pos "..minetest.pos_to_string(obj:get_pos(),0)..
			" on command of "..p_name)
	end
end

-- has the player the right privs?
-- this is used for the "I am your master" talk based configuration; *NOT* for the staffs!
yl_speak_up.may_edit_npc = function(player, n_id)
	if(not(player)) then
		return false
	end
	local pname = player:get_player_name()
	-- is the player allowed to edit this npc?
	return ((yl_speak_up.npc_owner[ n_id ] == pname
	  and minetest.check_player_privs(player, {npc_talk_owner=true}))
	  or minetest.check_player_privs(player, {npc_talk_master=true})
	  or minetest.check_player_privs(player, {npc_master=true})
	  or yl_speak_up.speak_to[pname].may_edit_this_npc)
end


-- log changes done by players or admins to NPCs
yl_speak_up.log_change = function(pname, n_id, text)
	-- make sure all variables are defined
	if(not(pname)) then
		pname = "- unkown player -"
	end
	if(not(n_id)) then
		n_id = "- unknown NPC -"
	end
	if(not(text)) then
		text = "- no text given -"
	end
	-- we don't want newlines in the texts
	text = string.gsub(text, "\n", "\\n")

	local log_text = "<"..tostring(n_id).."> ["..tostring(pname).."]: "..text
	-- log in general logfile
	minetest.log("yl_speak_up "..log_text)
	-- log with timestamp
	local log_text = tostring(os.time())..log_text
	-- TODO: log in a file for each npc and show it on demand?
end
